/*--------------------------------------------------------------------------
    Pawny 0.1.6, chess engine (source code).
    Copyright (C) 2009 - 2010 by Mincho Georgiev.
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    contact: pawnychess@gmail.com 
    web: http://www.pawny.netii.net/
----------------------------------------------------------------------------*/

#include "data.h"

#define ROOK_PAWN_CORRECTION 15
#define DOUBLED_PAWN_PENALTY 7 
#define ISOLATED_PAWN_PENALTY 5 
#define BACKWARDS_PAWN_PENALTY 8
#define PASSED_PAWN_REWARD 20
#define PAWN_CHAIN_ELEMENT 5
#define PROTECTED_PASSED_PAWN 50
#define HANGING_PAWN_PENALTY 2
#define PAWN_ON_7TH_REWARD 40

#define PHINDEX (board->phash % opt->pnt_numentry)

typedef uint64 bitboard_t;

//pawn table entry. (32 bytes of size)
typedef struct pte_t_ 
{	int score[2];
	bitboard_t wpassers;
	bitboard_t bpassers;
	hashkey_t  pkey;
}pte_t;

//pawn table pointer:
pte_t *pnt = 0;

//king near passed pawn bonus:
const int king_activity[8] = {0, 250, 200, 150, 100, 50, 30, 0};
const int passer_bonus[2][8] = {
	{0, 5, 10, 20, 30, 40, 50, 100},
	{100, 50, 40, 30, 20, 10, 5, 0}
};

//two tables for bit index conversion:
const int real_square_zip[128] = 
{
	 1,  2,  3,  4,  5,  6,  7,  8, 0,0,0,0,0,0,0,0,
	 9, 10, 11, 12, 13, 14, 15, 16, 0,0,0,0,0,0,0,0,
	17, 18, 19, 20, 21, 22, 23, 24, 0,0,0,0,0,0,0,0,
	25, 26, 27, 28, 29, 30, 31, 32, 0,0,0,0,0,0,0,0,
	33, 34, 35, 36, 37, 38, 39, 40, 0,0,0,0,0,0,0,0,
	41, 42, 43, 44, 45, 46, 47, 48, 0,0,0,0,0,0,0,0,
	49, 50, 51, 52, 53, 54, 55, 56, 0,0,0,0,0,0,0,0,
	57, 58, 59, 60, 61, 62, 63, 64, 0,0,0,0,0,0,0,0
};
const int real_square_unzip[65] = 
{0,
	  0,  1,  2,  3,  4,  5,  6,  7, 
	 16, 17, 18, 19, 20, 21, 22, 23, 
	 32, 33, 34, 35, 36, 37, 38, 39, 
	 48, 49, 50, 51, 52, 53, 54, 55, 
	 64, 65, 66, 67, 68, 69, 70, 71,
	 80, 81, 82, 83, 84, 85, 86, 87, 
	 96, 97, 98, 99,100,101,102,103, 
	112,113,114,115,116,117,118,119
};


//N.G. de Bruijn sequence for his multiplication method:
const bitboard_t magic = 0x022FDD63CC95386d;
const unsigned int magic_table[64] =
{
	 0,  1,  2, 53,  3,  7, 54, 27,
	 4, 38, 41,  8, 34, 55, 48, 28,
	62,  5, 39, 46, 44, 42, 22,  9,
	24, 35, 59, 56, 49, 18, 29, 11,
	63, 52,  6, 26, 37, 40, 33, 47,
	61, 45, 43, 21, 23, 58, 17, 10,
	51, 25, 36, 32, 60, 20, 57, 16,
	50, 31, 19, 15, 30, 14, 13, 12
};
 
void __inline pnt_free() { if(pnt) free(pnt); }
void __inline pnt_clear(){ memset(pnt, 0, (size_t)(opt->pnt_size));}
void __inline bitset(bitboard_t *bb, int index) { *bb |= (1ULL << index); }
void __inline bitclear(bitboard_t *bb, int index) {*bb &= ~(1ULL << index);}
int  __inline bitscanf(bitboard_t bb){
	return magic_table[((bb&-(signed long long)bb)*magic) >> 58];
}
	
//initializing the pawn hash table
int pnt_init(int MB)
{///min size - 8MB, max - 128:
	pnt_free();
	if(MB < 8) MB = 8; if(MB > 128) MB = 128;
	opt->pnt_size = (MB*1024*1024);
	opt->pnt_numentry = opt->pnt_size / sizeof(pte_t);	
	pnt = (pte_t *)malloc((size_t)opt->pnt_size);
	if(!pnt) return false;
	memset(pnt, 0,(size_t)(opt->pnt_size));
	return true;
}


static bool pnt_probe(bitboard_t *wpassers, bitboard_t *bpassers,int score[])
{
	pte_t *p = &pnt[PHINDEX];
	if(p->pkey == board->phash)
	{
		*wpassers = p->wpassers;
		*bpassers = p->bpassers;
		score[W] = p->score[W];
		score[B] = p->score[B];
		return true;
	}
	return false;
}


static void pnt_save(bitboard_t wpassers, bitboard_t bpassers, int score[])
{
	pte_t *p = &pnt[PHINDEX];
	p->pkey = board->phash;
	p->wpassers = wpassers;
	p->bpassers = bpassers;
	p->score[W] = score[W];
	p->score[B] = score[B];
}


static int eval_white_passer(int p_sq)
{
	int i,file,rank;
	int score = 0;
	int king_sq = King_Square(W);
	int op_king_sq = King_Square(B);
	
	file = calc_file(p_sq) + 1;
	rank = calc_rank(p_sq);
	
	score += passer_bonus[W][rank];
	//adding score for king near that pawn
	//and penalty for opking being near:
	score += (king_activity[distance_table[king_sq][p_sq]] >> 2);
	score -= (king_activity[distance_table[op_king_sq][p_sq]] >> 2);
			
	//is there a friendly pawn behind diagonally:
	if(pawn_first[WHITE][file - 1] < rank)
		//adding bonus, dependig of how far the protector is
		score += (PROTECTED_PASSED_PAWN >> (rank - pawn_first[WHITE][file - 1]));
						
	//the other side	
	if(pawn_first[WHITE][file + 1] < rank)
		//adding bonus, dependig of how far the protector is
		score += (PROTECTED_PASSED_PAWN >> (rank - pawn_first[WHITE][file - 1]));
				
	//if it's backed up by a rook:
	for(i = p_sq; calc_rank(i) != RANK_1-1; i-=16)
		if(PieceType(i) == WHITE_ROOK) score += 15;
	
	return score;
}


static int eval_black_passer(int p_sq)
{
	int i,file,rank;
	int score = 0;
	int king_sq = King_Square(B);
	int op_king_sq = King_Square(W);
	
	file = calc_file(p_sq) + 1;
	rank = calc_rank(p_sq);
	
	score += passer_bonus[B][rank];
	//adding score for king near that pawn
	//and penalty for opking being near:
	score += (king_activity[distance_table[king_sq][p_sq]] >> 2);
	score -= (king_activity[distance_table[op_king_sq][p_sq]] >> 2);

	//is there a friendly pawn behind diagonally:
	if(pawn_first[BLACK][file - 1] > rank)
		//adding bonus, dependig of how far the protector is
		score += (PROTECTED_PASSED_PAWN >> (pawn_first[BLACK][file - 1] - rank));
	//the other side:
	if(pawn_first[BLACK][file + 1] > rank)
		//adding bonus, dependig of how far the protector is
		score += (PROTECTED_PASSED_PAWN >> (pawn_first[BLACK][file + 1] - rank));

	//if it's backed up by a rook:
	for(i = p_sq; calc_rank(i) != RANK_8+1; i+=16)
		if(PieceType(i) == BLACK_ROOK) score += 15;

	return score;
}

static int eval_white_pawn(int sq)
{
	int score = 0;
	int file = calc_file(sq) + 1;
	int rank = calc_rank(sq);
	
	//the rook pawn costs 15% less than the others:
	if((file - 1) == FILE_A || (file - 1) == FILE_H)
		score -= ROOK_PAWN_CORRECTION;
	
	if(PieceType(sq - 15) == WP) score += PAWN_CHAIN_ELEMENT;
	if(PieceType(sq - 17) == WP) score += PAWN_CHAIN_ELEMENT;
	
	//pawn on 7th
	if(rank == RANK_7) score += PAWN_ON_7TH_REWARD;
	//doubled pawn:
	if(pawn_first[WHITE][file] < rank)
		score -= DOUBLED_PAWN_PENALTY;

	//isolated pawn:
	if(!pawn_file[WHITE][file - 1] 
	&& !pawn_file[WHITE][file + 1])
		score -= ISOLATED_PAWN_PENALTY;

	//hanging pawns (isolated / 4 - because the sum of both will be isolated / 2)
	//starting with the left hanging pawn
	if(!pawn_file[WHITE][file - 1]
	&& pawn_file[WHITE][file + 1])
		score -= HANGING_PAWN_PENALTY;
	//the right hanging pawn:
	if(!pawn_file[WHITE][file + 1]
	&& pawn_file[WHITE][file - 1])
		score -= HANGING_PAWN_PENALTY;
	//backward pawns:
	if(pawn_first[WHITE][file - 1] > rank
	&& pawn_first[WHITE][file + 1] > rank)
		score -= BACKWARDS_PAWN_PENALTY;
	
	return score;
}

static int eval_black_pawn(int sq)
{
	int score = 0;
	int file = calc_file(sq) + 1;
	int rank = calc_rank(sq);
	
	//the rook pawn costs 15% less than the others:
	if((file - 1) == FILE_A || (file - 1) == FILE_H)
		score -= ROOK_PAWN_CORRECTION;
	
	if(PieceType(sq + 15) == BP) score += PAWN_CHAIN_ELEMENT;
	if(PieceType(sq + 17) == BP) score += PAWN_CHAIN_ELEMENT;
	
	//pawn on 7th
	if(rank == RANK_2) score += PAWN_ON_7TH_REWARD;
		
	if(pawn_first[BLACK][file] > rank) 
		score -= DOUBLED_PAWN_PENALTY;

	if(!pawn_file[BLACK][file - 1] 
	&& !pawn_file[BLACK][file + 1])
		score -= ISOLATED_PAWN_PENALTY;

		
	if(!pawn_file[BLACK][file - 1]
	&& pawn_file[BLACK][file + 1])
		score -= HANGING_PAWN_PENALTY;
	
	if(!pawn_file[BLACK][file + 1]
	&& pawn_file[BLACK][file - 1])
		score -= HANGING_PAWN_PENALTY;
		
	if(pawn_first[BLACK][file - 1] < rank
	&& pawn_first[BLACK][file + 1] < rank)
		score -= BACKWARDS_PAWN_PENALTY;
	
	return score;
}


void eval_pawn_struct(int score[])
{	
	int i,sq,x,file,rank;
	bitboard_t wp = 0;
	bitboard_t bp = 0;
	int passer_score[2] = {0};

	for(i = 0; i < 10; i++)
	{	pawn_file[WHITE][i] = 0;
		pawn_file[BLACK][i] = 0;
		pawn_first[WHITE][i] = RANK_7;
		pawn_first[BLACK][i] = RANK_2;
	}
	///collecting PAWN information:
	for(sq = PLS(WP); sq <= H8; sq = PLN(sq))
	{//white pawns:
		file = calc_file(sq) + 1;
		rank = calc_rank(sq);
		pawn_file[WHITE][file]++;
		if(pawn_first[WHITE][file] > rank)
			pawn_first[WHITE][file] = rank;
	}
	for(sq = PLS(BP); sq <= H8; sq = PLN(sq))
	{//black pawns
		file = calc_file(sq) + 1;
		rank = calc_rank(sq);
		pawn_file[BLACK][file]++;
		if(pawn_first[BLACK][file] < rank)
			pawn_first[BLACK][file] = rank;
	}
	
	///probe the pawn table:
	if(pnt_probe(&wp,&bp,&score[0]))
	{
		do{ //examine white passers
			x = bitscanf(wp);
			bitclear(&wp, x);
			sq = real_square_unzip[x];
			if(sq)
			{	if(endgame) passer_score[W] += eval_white_passer(sq);
				else passer_score[W] += PASSED_PAWN_REWARD;
			}
		}while(x);
		do{ ////examine black passers
			x = bitscanf(bp);
			bitclear(&bp, x);
			sq = real_square_unzip[x];
			if(sq)
			{	if(endgame) passer_score[B] += eval_black_passer(sq);
				else passer_score[B] += PASSED_PAWN_REWARD;
			}
		}while(x);
	}
	else
	{
		///pawn evaluation
		for(sq = PLS(WP); sq <= H8; sq = PLN(sq))
		{
			if(endgame) score[W] += endgame_pawn_psq[W][sq];
			else score[W] += psq_table[WP][sq];
			score[W] += eval_white_pawn(sq);
			
			//passed pawn
			file = calc_file(sq) + 1;
			rank = calc_rank(sq);
			if(pawn_first[BLACK][file - 1] <= rank
			&& pawn_first[BLACK][file + 1] <= rank
			&& !pawn_file[BLACK][file])
			{	if(endgame) passer_score[W] += eval_white_passer(sq);
				else passer_score[W] += PASSED_PAWN_REWARD;
				//saving passer square:
				bitset(&wp, real_square_zip[sq]);
			}
		}
		for(sq = PLS(BP); sq <= H8; sq = PLN(sq))
		{
			if(endgame) score[B] += endgame_pawn_psq[B][sq];
			else score[B] += psq_table[BP][sq];
			score[B] += eval_black_pawn(sq);
			
			//passed pawn:
			file = calc_file(sq) + 1;
			rank = calc_rank(sq);
			if(pawn_first[WHITE][file - 1] >= rank
			&& pawn_first[WHITE][file + 1] >= rank
			&& !pawn_file[WHITE][file])
			{	if(endgame) passer_score[B] += eval_black_passer(sq);
				else passer_score[B] += PASSED_PAWN_REWARD;
				bitset(&bp, real_square_zip[sq]);
			}
		}
		//save info to pawn table:
		pnt_save(wp, bp, &score[0]);
	}
	score[W] += passer_score[W]; score[B]+=passer_score[B];
}
