/*--------------------------------------------------------------------------
    Pawny 0.1.6, chess engine (source code).
    Copyright (C) 2009 - 2010 by Mincho Georgiev.
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    contact: pawnychess@gmail.com 
    web: http://www.pawny.netii.net/
----------------------------------------------------------------------------*/

#include "data.h"

#define FUTILITY_MARGIN_1 (N_VALUE + P_VALUE)
#define FUTILITY_MARGIN_2 (R_VALUE + P_VALUE + P_VALUE)
#define FUTILITY_MARGIN_3 (Q_VALUE)
#define FULLDEPTHMOVES 4
#define REDUCTION (2*PLY)
#define DTC_FACT (2.5)

static void search_init()
{	
	si->nodes = 0;
	si->stop_search = false;
	si->null_line = false;
	board->ply = 0;
	memset(si->pv,0,sizeof(si->pv));
	memset(si->pvlen,0,sizeof(si->pvlen));
	memset(si->history, 0, sizeof(si->history));
	memset(si->recapture,-1,sizeof(si->recapture));
	si->num_hash_saves = 0;
	tt_check();
}

static bool is_draw()
{
	int i;	
	if(board->mr50 >= 100) return true;
	for(i = board->hply - 1; i >= 0; i--){
		if(board->hs[i].hash == board->hash)
			return true;
	}
	return false;
}

static int adapt_R(int depth)
{
	knowledge_t *know = &board->know;
	if(depth > (6*PLY))
	{	if(know->material - \
		(know->pawn_value[WHITE] + \
		know->pawn_value[BLACK]) > 4*R_VALUE)
			return (3*PLY);
		return (2*PLY);
	}
	else return (2*PLY);
}

static void pick_move(move_t ms[],int count, int index)
{
	int i,best_score = -INF,
	best_index = index;
	move_t temp;
	
	for(i = index; i < count; i++)
	{	if(ms[i].score > best_score)
		{	best_score = ms[i].score;
			best_index = i;
		}
	}	
	temp = ms[index];
	ms[index] = ms[best_index];
	ms[best_index] = temp;
}

static void update_pv(move_t move)
{
	register int i,ply = board->ply;	
	for (i = ply + 1;  i < si->pvlen[ply + 1]; i++)
		si->pv[ply][i] = si->pv[ply + 1][i];
	si->pvlen[ply] = si->pvlen[ply+1];
	si->pv[ply][ply].p = move.p;
	si->pv[ply][i].p = 0;
}

static void update_killers(move_t move)
{
	int ply = board->ply;
	if((move.type & CAP) || (move.type & PROM)) return;
	if(move.p == si->killer[0][ply]) return;
	si->killer[1][ply] = si->killer[0][ply];
	si->killer[0][ply] = move.p;
}

static int ok_to_reduce(move_t m)
{
	int ply = board->ply;
	if((m.type & CAP) || (m.type & PROM)
	|| pawn_endgame || king_alone)
		return false;
	if(m.p == si->killer[0][ply])
		return false;
	if(m.p == si->killer[1][ply])
		return false;
	return true;
}

static void checkup()
{
	if(opt->mode == UCI_MODE) check_for_input();
	if(!opt->analyze) time_check();
}

static int qsearch(int alpha, int beta, bool pv)
{
	int i,value,incheck,delta;
	int move_count;
	move_t ms[MOVE_STACK];
	int legal_moves = 0;
	
	if(!(si->nodes & 0x1fff)) checkup();
	if(si->stop_search) return 0;
	si->nodes++;
	si->pvlen[board->ply] = board->ply;
	
	value = eval();
	if(value >= beta) return value;
	if(value > alpha) alpha = value;
	
	incheck = is_in_check(board->side);
	if(!incheck) incheck = is_in_check(board->xside);
	if(incheck)
	{	move_count = move_gen(&ms[0]);
		if(pv) ordering(&ms[0], move_count, si->pv[0][board->ply].p);
		else   ordering(&ms[0], move_count, 0);
	}
	else
	{	move_count = move_gen_caps(&ms[0]);
		caps_ordering(&ms[0],move_count);
	}
	delta = alpha - P_VALUE - value;
	for(i = 0; i < move_count;i ++)
	{	pick_move(&ms[0],move_count, i);
		//delta pruning:
		if(!incheck && !ms[i].promoted
		&& (ms[i].score < delta)) continue;
		if(!move_make(ms[i])) continue;
		legal_moves++;
		
		if(legal_moves == 1) value = -qsearch(-beta,-alpha,pv);
		else value = -qsearch(-beta,-alpha,0);
		
		move_undo();
		if(si->stop_search) break;
		
		if(value > alpha)
		{	update_pv(ms[i]);
			if(value >= beta) return value;
			alpha = value;
		}
	}
	return alpha;
}

static int search(int alpha, int beta, int depth, bool null_ok, bool pv)
{
	int i,value,move_count;
	nullstack_t ns;
	move_t ms[MOVE_STACK];
	int ply = board->ply,fmb, fut_value = 0;
	sint16 ext = 0;
	bool incheck, givescheck;
	bool mate_threat = false;
	bool pvnode = (bool)((beta - alpha) > 1);
	uint8 val_flag = TT_ALPHA;
	uint32 best_move = 0;
	int legal_moves = 0, searched_moves = 0;
	bool futility_ok = false;
	knowledge_t *know = &board->know;
		
	if(alpha > MATE_VALUE - ply - 1) return alpha;
	if(ply && is_draw()) return 0;
	if(!(si->nodes & 0x1fff)) checkup();
	if(si->stop_search) return 0;
	si->pvlen[ply] = ply;

	switch(tt_probe(depth,&value, &best_move, &null_ok))
	{	case TT_ALPHA: 	if(value <= alpha) return alpha; 
				if(value < beta) beta = value; break;
		case TT_BETA:	if(value >= beta) return value; 
				if(value > alpha) alpha = value; break;
		case TT_EXACT:	return value;
		default:break;
	}

	if(depth < PLY) return qsearch(alpha,beta,pv);
	si->nodes++;
	incheck = is_in_check(board->side);
	
	//nullmove
	if(!incheck && null_ok)
	{	int R = adapt_R(depth);
		if(depth > R && !pawn_endgame && !king_alone)
		{	si->null_line = true;
			move_make_null(&ns);
			value = -search(-beta, -beta + 1, depth - R - PLY, false, 0);
			move_undo_null(&ns);
			si->null_line = false;
			if(value >= beta) return value;
			if(value < (-MATE_VALUE + MAX_PLY))
				mate_threat = true;
		}
	}
	
	//check and mate threat extensions:
	ext += incheck * (PLY);
	ext += mate_threat * (PLY/2);
	
	if(incheck)
	{	if((move_count = move_gen_legal(&ms[0])) == 1)
			ext += (PLY);//singular reply extension
	}
	else move_count = move_gen(&ms[0]);
		
	if(pv && si->pv[0][ply].p) best_move = si->pv[0][ply].p;
	//Internal Iterative Deepening:
	if(!best_move && pvnode && depth >= (3*PLY) 
	&& !pawn_endgame && null_ok && !si->null_line)
	{	value = search(alpha,beta,depth - (2*PLY),true, 0);
		if(value <= alpha)
			search(-INF,beta,depth - (2*PLY),true, 0);
		best_move = si->pv[ply][ply].p;
	}
	
	//determine the futility value:
	if(!pvnode && !ext)
	{	if(depth == (3*PLY))
		{	fmb = full_material_ballance(board->side);
			if((fmb + FUTILITY_MARGIN_3) <= alpha 
			&& (know->piece_count[board->xside] > 3)) 
				depth -= (PLY);
		}
		if(depth == (2*PLY))
		{	fmb = full_material_ballance(board->side);
			if(fmb + FUTILITY_MARGIN_2 <= alpha)
			{	futility_ok = true;
				fut_value = fmb + FUTILITY_MARGIN_2;
			}
		}
		if(depth == (PLY))
		{	fmb = full_material_ballance(board->side);
			if(fmb + FUTILITY_MARGIN_1 <= alpha)
			{	futility_ok = true;
				fut_value = fmb + FUTILITY_MARGIN_1;
			}
		}
	}
	ordering(&ms[0],move_count,best_move);
	for(i = 0; i < move_count; i++)
	{	givescheck = false;
		pick_move(&ms[0],move_count, i);
		if(!move_make(ms[i])) continue;
		givescheck = is_in_check(board->side);
		//recapture extension:
		if((ms[i].type & CAP) && pvnode)
		{	int pmax = max(pval[PieceType(ms[i].to)], \
				pval[board->hs[board->hply-1].moved]);
			int pmin = min(pval[PieceType(ms[i].to)], \
				pval[board->hs[board->hply-1].moved]);
			if(si->recapture[ply-1] == ms[i].to){	
				depth += (PLY/2);
				si->recapture[ply-1] = -1;
			}
			if(pmin >= N_VALUE && pmax >= N_VALUE){
				if(pmax - pmin <= (B_VALUE-N_VALUE))
					si->recapture[ply] = ms[i].to;
			}
		}
		//queen promotion extension:
		if(ms[i].promoted == Coloured(QUEEN)) ext += (PLY/2);
		//pawn push extension:
		if(GetType(PieceType(ms[i].to)) == PAWN)
		{	if(GetColor(PieceType(ms[i].to)) == WHITE
			&& calc_rank(ms[i].to == RANK_7))
				ext += (PLY/2);
			if(GetColor(PieceType(ms[i].to)) == BLACK
			&& calc_rank(ms[i].to == RANK_2))
				ext += (PLY/2);
		}
		//futility pruning:
		if(futility_ok && !ext && !givescheck)
		{	if(fut_value + material_gain(ms[i]) <= alpha)
			{	move_undo();
				legal_moves++;
				continue;
			}
		}
		//LMR:
		if(legal_moves && !ext && ok_to_reduce(ms[i]) && !givescheck)
		{	if(legal_moves > FULLDEPTHMOVES
			&& depth > REDUCTION)
				value = -search(-alpha - 1, -alpha, depth - REDUCTION,true, 0);
			else value = alpha + 1;
			if(value > alpha)
			{ 	value = -search(-alpha - 1,-alpha,depth - PLY,true, 0);
				if((value > alpha) && (value < beta))
					value = -search(-beta,-alpha,depth - PLY,true, 0);
			}
			move_undo();
			goto rec_value;
		}
		//PVS:
		if(!legal_moves)
			value = -search(-beta,-alpha,depth - PLY + ext,true,pv);
		else 
		{	value = -search(-alpha - 1,-alpha,depth - PLY + ext,true,0);
			if((value > alpha) && (value < beta))
				value = -search(-beta,-alpha,depth - PLY + ext,true,0);
		}
		move_undo();
		if(si->stop_search) break;
		rec_value:
		legal_moves++;
		searched_moves++;
		if(value > alpha){	
			best_move = ms[i].p;
			update_pv(ms[i]);
			if(value >= beta)
			{	update_killers(ms[i]);
				tt_save(depth,value,TT_BETA,best_move,null_ok);
				return value;
			}
			alpha = value;
			val_flag = TT_EXACT;
			si->history[ms[i].from][ms[i].to] += (depth/PLY);
		}
	}
	if(!legal_moves)
	{	if(incheck) return (-MATE_VALUE + ply); //checkmate
		else return STALEMATE_VALUE; //stalemate
	}
	//try for "all-moves cut"
	if(legal_moves && !searched_moves) return qsearch(alpha,beta,0);
	//transposition update if it's not a mate or stalemate!
	tt_save(depth,alpha,val_flag,best_move,null_ok);
	return alpha;
}

static int root_search(int alpha,int beta,int depth)
{
	int i,value;
	move_t ms[MOVE_STACK];
	int move_count;
	bool incheck;
	
	si->nodes++;
	si->root_moves = 0;
	si->pvlen[board->ply] = board->ply;
	incheck = is_in_check(board->side);
	move_count = move_gen(&ms[0]);
	ordering(&ms[0],move_count,si->pv[0][0].p);
	for(i = 0; i < move_count; i++)
	{	pick_move(&ms[0],move_count, i);
		if(!move_make(ms[i])) continue;
		si->root_moves++;
		if(opt->mode == UCI_MODE)
		{	printf("info currmove ");
			print_move(ms[i].p);
			printf(" currmovenumber %d\n", si->root_moves);
		}
		if(si->root_moves == 1)
			value = -search(-beta,-alpha,depth - PLY,true, 1);
		else 
		{	value = -search(-alpha-1,-alpha,depth - PLY,true, 0);
			if((value > alpha) && (value < beta))
				value = -search(-beta,-alpha,depth - PLY,true, 0);
		}
		move_undo();
		if(si->stop_search) break;
		if(value > alpha)
		{	update_pv(ms[i]);
			print_info(value, (depth/PLY));
			if(value >= beta)
			{	update_killers(ms[i]);
				return value;
			}
			alpha = value;
			si->history[ms[i].from][ms[i].to] += (depth/PLY);
		}
	}
	if(!si->root_moves)
	{	if (incheck) 
			return (-MATE_VALUE + board->ply);
		else return STALEMATE_VALUE;
	}
	return alpha;
}

void think()
{
	int i,depth,r = 0;
	float t1;
	int milliseconds,kns;
	move_t ponder,pvmove;
	uint64 total_nodes = 0; //used for calc. of the Kn/s
	int alpha = -INF;
	int beta = INF;
	uint64 last_iteration_time = 0;
	uint64 time_now = 0;
	uint64 time_prev = 0;

	//getting the initial time
	milliseconds = get_time();
	depth = opt->max_depth;
	if(!depth) depth = MAX_DEPTH;
	
	search_init();
	opt->startup_time = get_time();
	
	//book move:
	if(opt->book_active)
	{ if(book_move(&pvmove))
	  { if(move_make(pvmove))
	    { if(opt->mode == UCI_MODE)
	      { printf("bestmove ");
		print_move(pvmove.p);
		printf("\n");
		move_undo();
	      }
	      else
	      { printf("\nmove ");
		print_move(pvmove.p);
		printf("\n\n");
	      }
	      return;
	    }
	  }
	}
	//iterative deepening
	for(i = 1 ; i <= depth; i++)
	{	if(opt->mode == CONSOLE_MODE)
		{	total_nodes += si->nodes;
			si->nodes = 0;
		}

		if(opt->dtc) time_prev = get_time();
			
		r = root_search(alpha,beta,(i * PLY));
		
		if(opt->dtc && !opt->analyze){
			time_now = get_time();
			last_iteration_time = (time_now - time_prev);
			if((opt->max_time - (time_now - opt->startup_time)) < \
			(abs((int)(last_iteration_time * DTC_FACT))))
				break;
		}
		if(si->stop_search) break;
		
		if(r <= alpha || r >= beta)
		{	r = root_search(-INF,INF,(i * PLY));
			if(si->stop_search) break;
		}
			
		alpha = r - opt->aspiration;
		beta =  r + opt->aspiration;
		if(alpha < -INF) alpha = -INF;
		if(beta > INF) beta = INF; 
	}
	if(opt->mode == UCI_MODE)
	{	printf("bestmove ");
		print_move(si->pv[0][0].p);
		if(si->pv[0][1].p)
		{	printf("ponder ");
			print_move(si->pv[0][1].p);
		}
		else
		{	if(tt_retrieve_ponder(&ponder))
			{	printf("ponder ");
				print_move(ponder.p);
			}
		}
		printf("\n");
	}
	if(opt->mode == CONSOLE_MODE)
	{	pvmove.p = si->pv[0][0].p;
		//time and kilonodes per second calculations:
		t1 = ((get_time() - milliseconds) / (float)1000); //seconds
		kns = (int)((total_nodes/1000)/t1);
		if(kns > 0) printf ("   %d Kn/s,  %.2f sec\n",kns,t1);
		//mate or stalemate:
		if(r >= MATE_VALUE - MAX_PLY)
		{	if(r == MATE_VALUE - 1)
			{	move_make(pvmove);
				printf("\nmove ");
				print_move(pvmove.p);
				printf("\n\n");
				printf("\ncheckmate.\n\n");
				return;
			}
			else
			{	move_make(pvmove);
				printf("\nmove ");
				print_move(pvmove.p);
				printf("\n\n");
				printf("\nmate in %d.\n\n",((MATE_VALUE - r) / 2)+1);
				return;
			}
		}
		if(r == STALEMATE_VALUE && !si->root_moves)
		{	printf("\nstalemate.\n\n");
			return;
		}
		//making the best move and return
		move_make(pvmove);
		printf("\nmove ");
		print_move(pvmove.p);
		printf("\n\n");
	}
}
